/*
    Gaudium  - an HTTPSERVER for Nodejs
    Copyright (C)  2011  freefeathers | Tierracycle Cooperative

    Contributors:
                   Nathaniel Jayme <nbjayme@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

var _libhttp = require('http')
var _libfs = require('fs')
var _libpath = require('path')
var _libMime = require('./mime.js')

try
{
   var _libuuid = require('node-uuid')
}
catch(e)
{
   var _libuuid = require('../node-uuid')
}

var Worker = {}

Worker.HttpRemoteClient = require('./httpRemoteClient') //create each connection
Worker.Config = require('./defaultConfig')


/* singleton */
Worker.Util ={
    clearRequireCache: function (libName)
    {
          var _lib = libName.match(/[^/]+$/)
          if ( _lib == null )
          {
              return
          }

          _lib = _lib[0]

          var r = new RegExp( '/' + _lib + '$')
          for (var s in require.cache)
          {
               if ( r.exec(s) )
               {
                   delete require.cache[s]
               }
          }
    },

    getErrFileContent:function(basePath,code)
    {
          var errFile = basePath + "/" + Worker.Config.errDocs + "/doc" + code + ".html"
          return this.getFileContent(errFile)
    },

    getFileContent:function( fPath )
    {
          var data = null
          try
          {
              var stats = _libfs.statSync(fPath)
              if (stats.isFile() )
              {
                  data = _libfs.readFileSync(fPath,"UTF-8")
              }
          }
          catch (e){  }
          return data
    }
}

/* singleton */
Worker.State = {
     _isRunning:false,

    isRunning:function(flag)
    {
         if ( flag != null )
         {
               this._isRunning = flag
         }
         return this._isRunning
    }
}


Worker.HttpConnection =function(req,resp){
    // protected variables
     var _self = this
     var _id =   _libuuid()
     var _client = new Worker.HttpRemoteClient( req, resp )
     var _message = null 
     var _docRoot = null // to be assigned

     var host = _client.getHost()

     if ( host == '127.0.0.1' )
     {
            host = 'localhost'
     }

     if (  host == 'localhost' )
     {
            _docRoot  = Worker.Config.hosts[ "default" ].docRoot
     }

     if (  host in Worker.Config.hosts )
     {
            _docRoot = _libpath.normalize( Worker.Config.hosts[ this._host ].docRoot )
     }

     // public methods and properties
     this.freeMem=function()
     {
           _self = null
           _client = null
           delete _client
     }

     this.getClient = function()
     {
           return _client
     }

     this.getId = function()
     {
          return _id
     }
     this.getSelf = function()
     {
          return _self
     }

     this.getDocRoot = function()
     {
          return _docRoot
     }

     this._status = "onhold"
}

Worker.HttpConnection.prototype={
    getHost:function()
    {
          return this.getClient().getHost()
    },
    getPort:function()
    {
          return this.getClient().getPort()
    },
    getRequestChannel:function(){
        return this.getClient().getRequestChannel()
    },
    getResponseChannel:function(){
        return this.getClient().getResponseChannel()
    },
    getRequestURI:function(){
         return this.getClient().getRequestURI()
    },
    getStatus:function(){
        return this._status
    },
    setStatus:function( status )
    {
        this._status = status
    },
    sendOk:function( msgData, content_type )
    {
          var msgStruc = {
                "code": 200,
                "header": {
                      "content-type" : content_type
                },
                "data": msgData
          }
          this.getClient().sendMessage( msgStruc)
          this.setStatus( 'done');
    },
    sendNotFound:function()
    {
          var msgData = Worker.Util.getErrFileContent(this.getDocRoot(),404)
          if (msgData == null )
          {
              msgData = "Resource not found" 
          }

          var msgStruc = {
                'code': 404,
                'header': {
                      "content-type" : 'text/html'
                },
                'data': msgData
          }
          this.getClient().sendMessage( msgStruc)
          this.setStatus( 'done');
    },
    sendFileContent:function(fPath, content_type )
    {
           var _self = this.getSelf()
           var extension = _libpath.extname(fPath)
           if (extension) extension = extension.substr(1)
           if (extension == Worker.Config.nodeScript )
           {
                   setTimeout( function() {  Worker.Events.ExecuteScript(_self,fPath)  }, 150 )
                   return
           }

           if (extension in _libMime.Mime )
           {
                var mime = _libMime.Mime[ extension ]
           }

           if (content_type == null )
           {
               content_type = mime
           }
           
           if (content_type.match(/^text/) == null )
           {
              var msgData = _libfs.readFileSync(fPath)
           }
           else
           {
              var msgData = _libfs.readFileSync(fPath,'UTF-8')
           }

           var msgStruc = {
                'code': 200,
                'header': {
                      "content-type" : content_type
                },
                'data': msgData
           }
           this.getClient().sendMessage( msgStruc)
           this.setStatus( 'done');
    },
    sendUnhandledException:function(e)
    {
         var msgData = e.stack.replace('\n', '<br />')
         var dataHtml = Worker.Util.getErrFileContent(this.getDocRoot(),"UnhandledException")
         if ( dataHtml )
         {
               msgData = dataHtml +  "<br /> <br />Stack trace resut: <br /> <br />" + msgData
         }

         var msgStruc = 
         {
             'code': 200,
             'header': { 'content-type': "text/html" },
             'data': msgData
         }
         this.getClient().sendMessage( msgStruc )
         this.setStatus( 'done');
    },
    isPathSecure:function(fullPath){
        var requestPath = _libpath.normalize( fullPath )
        var re = new RegExp( "^" + this.getDocRoot() )
        if (  re.exec( requestPath ) == null ) {
             return false
        }
        return true
    },
    Execute:function(){
         if ( this.getDocRoot() == null )
         {
            this.sendNotFound()
            return
         }

         this.setStatus( 'executing' )

         var requestURI  = this.getRequestURI()
         var fullPath = _libpath.normalize ( this.getDocRoot() + requestURI.pathname )

         if ( ! this.isPathSecure (fullPath ) )
         {
               this.sendNotFound()
               return
         }

          try {
               var stats = _libfs.statSync(fullPath)
          }
          catch(e)
          {
               this.sendNotFound()
               return
          }

          if ( stats.isDirectory() )
          {
              var docIndeces = Worker.Config.docIndex

              for (var nIndex = 0, maxIndex= docIndeces.length; nIndex < maxIndex; nIndex++ )
              {
                  var indexFile = fullPath + docIndeces[nIndex]
                  try
                  {
                     stats = _libfs.statSync(indexFile)
                     fullPath = indexFile
                     break
                  }
                  catch(e)
                  {
                      continue
                  }
              }

              if ( ! stats.isFile() )
              {
                  this.sendNotFound()
                 return
              }
          }

          if (stats.isFile() )
          {
              this.sendFileContent(fullPath)
              return
          }

          this.sendNotFound()
    }
}


/* singleton */
Worker.Manager=
{
     _workers:{},  // client connections
     _activeCount:0,
     _cleanerTick:1200,

     CreateAWorker:function(req,resp)
     {
          setTimeout( function() { Worker.Manager.CleanDoneWorkers() } ,this._cleanerTick )
          this._activeCount++
          var connection = new Worker.HttpConnection(req,resp)
          var id = connection.getId()

          this._workers[ id ] = connection
          return this._workers[ id ]
     },

     CleanAWorker:function( id )
     {
          var connection = this.getAWorker( id )
          if (connection == null )
          {
             return
          }
          this._workers[ id ].freeMem()
          delete this._workers[ id ]
          this._activeCount--
     },

     getAWorker:function( id )
     {
           if ( id in this._workers )
           {
                return this._workers[ id ]
           }
           return null
     },

     CleanDoneWorkers:function(id)
     {
           for (var id in this._workers )
           {
                oWorker = this._workers[ id ]
                if ( oWorker.getStatus() != 'done' ) 
                {
                   continue
                }
                this.CleanAWorker( id )
           }
           if (this._activeCount > 0 )
           {
              setTimeout( function() { Worker.Manager.CleanDoneWorkers() } , this._cleanerTick)
           }
     }

}

/* singleton */
Worker.Events={
    Request:function(req,resp)
    {
       var workerConnection =  Worker.Manager.CreateAWorker( req, resp )
       workerConnection.Execute()
    },

    Activate:function(config)
    {
           if(   Worker.State.isRunning() )
           {
                 return
           }

           Worker.State.isRunning( true )

            if ( config != null )
           {
                for (item in config)
                {
                     Worker.Config[item]  = config[item]
                }
           }

           console.log("Starting Web server at port " + Worker.Config.port )
           console.log("Hosted domains: " )
           for ( var h  in Worker.Config.hosts )
           {
                console.log ( ' domain ( ' + h + ')  -----> ' + Worker.Config.hosts[ h ].docRoot )
           }

           Worker.Service = new _libhttp.Server() 
           Worker.Service.listen(Worker.Config.port)
           Worker.Service.on("request", function(req,resp) {
                   setTimeout( function() { Worker.Events.Request(req,resp) }, 150)                              
            })
    },

    ExecuteScript:function( connection, fPath )
    {
        try
        {
           var nodeScript = require( fPath )
           Worker.Util.clearRequireCache( fPath )
           nodeScript.Execute( connection )
        }
        catch(e)
        {
            connection.sendUnhandledException( e )
            Worker.Util.clearRequireCache( fPath )
        }
        connection.setStatus( 'done');
    }

}

exports.GaudiumServer=function()
{
    obj = {
         isRunning:function()
         {
             return Worker.State.isRunning()
         },

         startService:function(config)
         {
             Worker.Events.Activate(config)
         }
   }
   return obj
}

